package org.apache.bcel.classfile;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.CharArrayReader;
import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Locale;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import org.apache.bcel.Const;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.classfile.ConstantPool;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.util.ByteSequence;
import org.apache.commons.lang3.ArrayUtils;

public abstract class Utility {
  private static final ThreadLocal<Integer> CONSUMER_CHARS = ThreadLocal.withInitial(() -> Integer.valueOf(0));
  
  private static boolean wide;
  
  private static final int FREE_CHARS = 48;
  
  private static final int[] CHAR_MAP = new int[48];
  
  private static final int[] MAP_CHAR = new int[256];
  
  private static final char ESCAPE_CHAR = '$';
  
  static {
    int j = 0;
    int i;
    for (i = 65; i <= 90; i++) {
      CHAR_MAP[j] = i;
      MAP_CHAR[i] = j;
      j++;
    } 
    for (i = 103; i <= 122; i++) {
      CHAR_MAP[j] = i;
      MAP_CHAR[i] = j;
      j++;
    } 
    CHAR_MAP[j] = 36;
    MAP_CHAR[36] = j;
    j++;
    CHAR_MAP[j] = 95;
    MAP_CHAR[95] = j;
  }
  
  public static String accessToString(int accessFlags) {
    return accessToString(accessFlags, false);
  }
  
  public static String accessToString(int accessFlags, boolean forClass) {
    StringBuilder buf = new StringBuilder();
    int p = 0;
    for (int i = 0; p < 32768; i++) {
      p = pow2(i);
      if ((accessFlags & p) != 0)
        if (!forClass || (p != 32 && p != 512))
          buf.append(Const.getAccessName(i)).append(" ");  
    } 
    return buf.toString().trim();
  }
  
  private static short byteToShort(byte b) {
    return (b < 0) ? (short)(256 + b) : (short)b;
  }
  
  public static String classOrInterface(int accessFlags) {
    return ((accessFlags & 0x200) != 0) ? "interface" : "class";
  }
  
  public static int clearBit(int flag, int i) {
    int bit = pow2(i);
    return ((flag & bit) == 0) ? flag : (flag ^ bit);
  }
  
  public static String codeToString(byte[] code, ConstantPool constantPool, int index, int length) {
    return codeToString(code, constantPool, index, length, true);
  }
  
  public static String codeToString(byte[] code, ConstantPool constantPool, int index, int length, boolean verbose) {
    StringBuilder buf = new StringBuilder(code.length * 20);
    try (ByteSequence stream = new ByteSequence(code)) {
      int i;
      for (i = 0; i < index; i++)
        codeToString(stream, constantPool, verbose); 
      for (i = 0; stream.available() > 0; i++) {
        if (length < 0 || i < length) {
          String indices = fillup(stream.getIndex() + ":", 6, true, ' ');
          buf.append(indices).append(codeToString(stream, constantPool, verbose)).append('\n');
        } 
      } 
    } catch (IOException e) {
      throw new ClassFormatException("Byte code error: " + buf.toString(), e);
    } 
    return buf.toString();
  }
  
  public static String codeToString(ByteSequence bytes, ConstantPool constantPool) throws IOException {
    return codeToString(bytes, constantPool, true);
  }
  
  public static String codeToString(ByteSequence bytes, ConstantPool constantPool, boolean verbose) throws IOException {
    int low, high, npairs, index, vindex, constant, match[], jumpTable[], offset, i;
    Constant c;
    int nargs, dimensions;
    short opcode = (short)bytes.readUnsignedByte();
    int defaultOffset = 0;
    int noPadBytes = 0;
    StringBuilder buf = new StringBuilder(Const.getOpcodeName(opcode));
    if (opcode == 170 || opcode == 171) {
      int remainder = bytes.getIndex() % 4;
      noPadBytes = (remainder == 0) ? 0 : (4 - remainder);
      for (int j = 0; j < noPadBytes; j++) {
        byte b;
        if ((b = bytes.readByte()) != 0)
          System.err.println("Warning: Padding byte != 0 in " + Const.getOpcodeName(opcode) + ":" + b); 
      } 
      defaultOffset = bytes.readInt();
    } 
    switch (opcode) {
      case 170:
        low = bytes.readInt();
        high = bytes.readInt();
        offset = bytes.getIndex() - 12 - noPadBytes - 1;
        defaultOffset += offset;
        buf.append("\tdefault = ").append(defaultOffset).append(", low = ").append(low).append(", high = ").append(high).append("(");
        jumpTable = new int[high - low + 1];
        for (i = 0; i < jumpTable.length; i++) {
          jumpTable[i] = offset + bytes.readInt();
          buf.append(jumpTable[i]);
          if (i < jumpTable.length - 1)
            buf.append(", "); 
        } 
        buf.append(")");
        return buf.toString();
      case 171:
        npairs = bytes.readInt();
        offset = bytes.getIndex() - 8 - noPadBytes - 1;
        match = new int[npairs];
        jumpTable = new int[npairs];
        defaultOffset += offset;
        buf.append("\tdefault = ").append(defaultOffset).append(", npairs = ").append(npairs).append(" (");
        for (i = 0; i < npairs; i++) {
          match[i] = bytes.readInt();
          jumpTable[i] = offset + bytes.readInt();
          buf.append("(").append(match[i]).append(", ").append(jumpTable[i]).append(")");
          if (i < npairs - 1)
            buf.append(", "); 
        } 
        buf.append(")");
        return buf.toString();
      case 153:
      case 154:
      case 155:
      case 156:
      case 157:
      case 158:
      case 159:
      case 160:
      case 161:
      case 162:
      case 163:
      case 164:
      case 165:
      case 166:
      case 167:
      case 168:
      case 198:
      case 199:
        buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readShort());
        return buf.toString();
      case 200:
      case 201:
        buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readInt());
        return buf.toString();
      case 21:
      case 22:
      case 23:
      case 24:
      case 25:
      case 54:
      case 55:
      case 56:
      case 57:
      case 58:
      case 169:
        if (wide) {
          vindex = bytes.readUnsignedShort();
          wide = false;
        } else {
          vindex = bytes.readUnsignedByte();
        } 
        buf.append("\t\t%").append(vindex);
        return buf.toString();
      case 196:
        wide = true;
        buf.append("\t(wide)");
        return buf.toString();
      case 188:
        buf.append("\t\t<").append(Const.getTypeName(bytes.readByte())).append(">");
        return buf.toString();
      case 178:
      case 179:
      case 180:
      case 181:
        index = bytes.readUnsignedShort();
        buf.append("\t\t").append(constantPool.constantToString(index, (byte)9)).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 187:
      case 192:
        buf.append("\t");
      case 193:
        index = bytes.readUnsignedShort();
        buf.append("\t<").append(constantPool.constantToString(index, (byte)7)).append(">").append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 183:
      case 184:
        index = bytes.readUnsignedShort();
        c = constantPool.getConstant(index);
        buf.append("\t").append(constantPool.constantToString(index, c.getTag())).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 182:
        index = bytes.readUnsignedShort();
        buf.append("\t").append(constantPool.constantToString(index, (byte)10)).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 185:
        index = bytes.readUnsignedShort();
        nargs = bytes.readUnsignedByte();
        buf.append("\t").append(constantPool.constantToString(index, (byte)11)).append(verbose ? (" (" + index + ")\t") : "").append(nargs).append("\t").append(bytes.readUnsignedByte());
        return buf.toString();
      case 186:
        index = bytes.readUnsignedShort();
        buf.append("\t").append(constantPool.constantToString(index, (byte)18)).append(verbose ? (" (" + index + ")\t") : "").append(bytes.readUnsignedByte()).append(bytes.readUnsignedByte());
        return buf.toString();
      case 19:
      case 20:
        index = bytes.readUnsignedShort();
        buf.append("\t\t").append(constantPool.constantToString(index, constantPool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 18:
        index = bytes.readUnsignedByte();
        buf.append("\t\t").append(constantPool.constantToString(index, constantPool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 189:
        index = bytes.readUnsignedShort();
        buf.append("\t\t<").append(compactClassName(constantPool.getConstantString(index, (byte)7), false)).append(">").append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 197:
        index = bytes.readUnsignedShort();
        dimensions = bytes.readUnsignedByte();
        buf.append("\t<").append(compactClassName(constantPool.getConstantString(index, (byte)7), false)).append(">\t").append(dimensions).append(verbose ? (" (" + index + ")") : "");
        return buf.toString();
      case 132:
        if (wide) {
          vindex = bytes.readUnsignedShort();
          constant = bytes.readShort();
          wide = false;
        } else {
          vindex = bytes.readUnsignedByte();
          constant = bytes.readByte();
        } 
        buf.append("\t\t%").append(vindex).append("\t").append(constant);
        return buf.toString();
    } 
    if (Const.getNoOfOperands(opcode) > 0)
      for (int j = 0; j < Const.getOperandTypeCount(opcode); j++) {
        buf.append("\t\t");
        switch (Const.getOperandType(opcode, j)) {
          case 8:
            buf.append(bytes.readByte());
            break;
          case 9:
            buf.append(bytes.readShort());
            break;
          case 10:
            buf.append(bytes.readInt());
            break;
          default:
            throw new IllegalStateException("Unreachable default case reached!");
        } 
      }  
    return buf.toString();
  }
  
  public static String compactClassName(String str) {
    return compactClassName(str, true);
  }
  
  public static String compactClassName(String str, boolean chopit) {
    return compactClassName(str, "java.lang.", chopit);
  }
  
  public static String compactClassName(String str, String prefix, boolean chopit) {
    int len = prefix.length();
    str = pathToPackage(str);
    if (chopit && str.startsWith(prefix) && str.substring(len).indexOf('.') == -1)
      str = str.substring(len); 
    return str;
  }
  
  public static String pathToPackage(String str) {
    return str.replace('/', '.');
  }
  
  public static String convertString(String label) {
    char[] ch = label.toCharArray();
    StringBuilder buf = new StringBuilder();
    for (char element : ch) {
      switch (element) {
        case '\n':
          buf.append("\\n");
          break;
        case '\r':
          buf.append("\\r");
          break;
        case '"':
          buf.append("\\\"");
          break;
        case '\'':
          buf.append("\\'");
          break;
        case '\\':
          buf.append("\\\\");
          break;
        default:
          buf.append(element);
          break;
      } 
    } 
    return buf.toString();
  }
  
  private static int countBrackets(String brackets) {
    char[] chars = brackets.toCharArray();
    int count = 0;
    boolean open = false;
    for (char c : chars) {
      switch (c) {
        case '[':
          if (open)
            throw new IllegalArgumentException("Illegally nested brackets:" + brackets); 
          open = true;
          break;
        case ']':
          if (!open)
            throw new IllegalArgumentException("Illegally nested brackets:" + brackets); 
          open = false;
          count++;
          break;
      } 
    } 
    if (open)
      throw new IllegalArgumentException("Illegally nested brackets:" + brackets); 
    return count;
  }
  
  public static byte[] decode(String s, boolean uncompress) throws IOException {
    byte[] bytes;
    try(JavaReader jr = new JavaReader(new CharArrayReader(s.toCharArray())); ByteArrayOutputStream bos = new ByteArrayOutputStream()) {
      int ch;
      while ((ch = jr.read()) >= 0)
        bos.write(ch); 
      bytes = bos.toByteArray();
    } 
    if (uncompress) {
      GZIPInputStream gis = new GZIPInputStream(new ByteArrayInputStream(bytes));
      byte[] tmp = new byte[bytes.length * 3];
      int count = 0;
      int b;
      while ((b = gis.read()) >= 0)
        tmp[count++] = (byte)b; 
      bytes = Arrays.copyOf(tmp, count);
    } 
    return bytes;
  }
  
  public static String encode(byte[] bytes, boolean compress) throws IOException {
    if (compress)
      try(ByteArrayOutputStream baos = new ByteArrayOutputStream(); GZIPOutputStream gos = new GZIPOutputStream(baos)) {
        gos.write(bytes, 0, bytes.length);
        gos.close();
        bytes = baos.toByteArray();
      }  
    CharArrayWriter caw = new CharArrayWriter();
    try (JavaWriter jw = new JavaWriter(caw)) {
      for (byte b : bytes) {
        int in = b & 0xFF;
        jw.write(in);
      } 
    } 
    return caw.toString();
  }
  
  static boolean equals(byte[] a, byte[] b) {
    int size;
    if ((size = a.length) != b.length)
      return false; 
    for (int i = 0; i < size; i++) {
      if (a[i] != b[i])
        return false; 
    } 
    return true;
  }
  
  public static String fillup(String str, int length, boolean leftJustify, char fill) {
    int len = length - str.length();
    char[] buf = new char[Math.max(len, 0)];
    Arrays.fill(buf, fill);
    if (leftJustify)
      return str + new String(buf); 
    return new String(buf) + str;
  }
  
  public static String format(int i, int length, boolean leftJustify, char fill) {
    return fillup(Integer.toString(i), length, leftJustify, fill);
  }
  
  public static String getSignature(String type) {
    StringBuilder buf = new StringBuilder();
    char[] chars = type.toCharArray();
    boolean charFound = false;
    boolean delim = false;
    int index = -1;
    for (int i = 0; i < chars.length; i++) {
      switch (chars[i]) {
        case '\t':
        case '\n':
        case '\f':
        case '\r':
        case ' ':
          if (charFound)
            delim = true; 
          break;
        case '[':
          if (!charFound)
            throw new IllegalArgumentException("Illegal type: " + type); 
          index = i;
          break;
        default:
          charFound = true;
          if (!delim)
            buf.append(chars[i]); 
          break;
      } 
    } 
    int brackets = 0;
    if (index > 0)
      brackets = countBrackets(type.substring(index)); 
    type = buf.toString();
    buf.setLength(0);
    for (int j = 0; j < brackets; j++)
      buf.append('['); 
    boolean found = false;
    for (int k = 4; k <= 12 && !found; k++) {
      if (Const.getTypeName(k).equals(type)) {
        found = true;
        buf.append(Const.getShortTypeName(k));
      } 
    } 
    if (!found)
      buf.append('L').append(type.replace('.', '/')).append(';'); 
    return buf.toString();
  }
  
  public static boolean isJavaIdentifierPart(char ch) {
    return ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9') || ch == '_');
  }
  
  public static boolean isSet(int flag, int i) {
    return ((flag & pow2(i)) != 0);
  }
  
  public static String[] methodSignatureArgumentTypes(String signature) throws ClassFormatException {
    return methodSignatureArgumentTypes(signature, true);
  }
  
  public static String[] methodSignatureArgumentTypes(String signature, boolean chopit) throws ClassFormatException {
    List<String> vec = new ArrayList<>();
    try {
      int index = signature.indexOf('(') + 1;
      if (index <= 0)
        throw new ClassFormatException("Invalid method signature: " + signature); 
      while (signature.charAt(index) != ')') {
        vec.add(typeSignatureToString(signature.substring(index), chopit));
        index += unwrap(CONSUMER_CHARS);
      } 
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid method signature: " + signature, e);
    } 
    return vec.<String>toArray(ArrayUtils.EMPTY_STRING_ARRAY);
  }
  
  public static String methodSignatureReturnType(String signature) throws ClassFormatException {
    return methodSignatureReturnType(signature, true);
  }
  
  public static String methodSignatureReturnType(String signature, boolean chopit) throws ClassFormatException {
    String type;
    try {
      int index = signature.lastIndexOf(')') + 1;
      if (index <= 0)
        throw new ClassFormatException("Invalid method signature: " + signature); 
      type = typeSignatureToString(signature.substring(index), chopit);
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid method signature: " + signature, e);
    } 
    return type;
  }
  
  public static String methodSignatureToString(String signature, String name, String access) {
    return methodSignatureToString(signature, name, access, true);
  }
  
  public static String methodSignatureToString(String signature, String name, String access, boolean chopit) {
    return methodSignatureToString(signature, name, access, chopit, (LocalVariableTable)null);
  }
  
  public static String methodSignatureToString(String signature, String name, String access, boolean chopit, LocalVariableTable vars) throws ClassFormatException {
    String type;
    StringBuilder buf = new StringBuilder("(");
    int varIndex = access.contains("static") ? 0 : 1;
    try {
      int index = signature.indexOf('(') + 1;
      if (index <= 0)
        throw new ClassFormatException("Invalid method signature: " + signature); 
      while (signature.charAt(index) != ')') {
        String paramType = typeSignatureToString(signature.substring(index), chopit);
        buf.append(paramType);
        if (vars != null) {
          LocalVariable l = vars.getLocalVariable(varIndex, 0);
          if (l != null)
            buf.append(" ").append(l.getName()); 
        } else {
          buf.append(" arg").append(varIndex);
        } 
        if ("double".equals(paramType) || "long".equals(paramType)) {
          varIndex += 2;
        } else {
          varIndex++;
        } 
        buf.append(", ");
        index += unwrap(CONSUMER_CHARS);
      } 
      index++;
      type = typeSignatureToString(signature.substring(index), chopit);
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid method signature: " + signature, e);
    } 
    if (buf.length() > 1)
      buf.setLength(buf.length() - 2); 
    buf.append(")");
    return access + (!access.isEmpty() ? " " : "") + type + " " + name + buf
      .toString();
  }
  
  public static String methodTypeToSignature(String ret, String[] argv) throws ClassFormatException {
    StringBuilder buf = new StringBuilder("(");
    if (argv != null)
      for (String element : argv) {
        String str1 = getSignature(element);
        if (str1.endsWith("V"))
          throw new ClassFormatException("Invalid type: " + element); 
        buf.append(str1);
      }  
    String str = getSignature(ret);
    buf.append(")").append(str);
    return buf.toString();
  }
  
  private static int pow2(int n) {
    return 1 << n;
  }
  
  public static String printArray(Object[] obj) {
    return printArray(obj, true);
  }
  
  public static String printArray(Object[] obj, boolean braces) {
    return printArray(obj, braces, false);
  }
  
  public static String printArray(Object[] obj, boolean braces, boolean quote) {
    if (obj == null)
      return null; 
    StringBuilder buf = new StringBuilder();
    if (braces)
      buf.append('{'); 
    for (int i = 0; i < obj.length; i++) {
      if (obj[i] != null) {
        buf.append(quote ? "\"" : "").append(obj[i]).append(quote ? "\"" : "");
      } else {
        buf.append("null");
      } 
      if (i < obj.length - 1)
        buf.append(", "); 
    } 
    if (braces)
      buf.append('}'); 
    return buf.toString();
  }
  
  public static void printArray(PrintStream out, Object[] obj) {
    out.println(printArray(obj, true));
  }
  
  public static void printArray(PrintWriter out, Object[] obj) {
    out.println(printArray(obj, true));
  }
  
  public static String replace(String str, String old, String new_) {
    try {
      if (str.contains(old)) {
        StringBuilder buf = new StringBuilder();
        int oldIndex = 0;
        int index;
        while ((index = str.indexOf(old, oldIndex)) != -1) {
          buf.append(str, oldIndex, index);
          buf.append(new_);
          oldIndex = index + old.length();
        } 
        buf.append(str.substring(oldIndex));
        str = buf.toString();
      } 
    } catch (StringIndexOutOfBoundsException e) {
      System.err.println(e);
    } 
    return str;
  }
  
  public static short searchOpcode(String name) {
    name = name.toLowerCase(Locale.ENGLISH);
    for (short i = 0; i < Const.OPCODE_NAMES_LENGTH; i = (short)(i + 1)) {
      if (Const.getOpcodeName(i).equals(name))
        return i; 
    } 
    return -1;
  }
  
  public static int setBit(int flag, int i) {
    return flag | pow2(i);
  }
  
  public static String signatureToString(String signature) {
    return signatureToString(signature, true);
  }
  
  public static String signatureToString(String signature, boolean chopit) {
    String type = "";
    String typeParams = "";
    int index = 0;
    if (signature.charAt(0) == '<') {
      typeParams = typeParamTypesToString(signature, chopit);
      index += unwrap(CONSUMER_CHARS);
    } 
    if (signature.charAt(index) == '(') {
      type = typeParams + typeSignaturesToString(signature.substring(index), chopit, ')');
      index += unwrap(CONSUMER_CHARS);
      type = type + typeSignatureToString(signature.substring(index), chopit);
      index += unwrap(CONSUMER_CHARS);
      return type;
    } 
    type = typeSignatureToString(signature.substring(index), chopit);
    index += unwrap(CONSUMER_CHARS);
    if (typeParams.isEmpty() && index == signature.length())
      return type; 
    StringBuilder typeClass = new StringBuilder(typeParams);
    typeClass.append(" extends ");
    typeClass.append(type);
    if (index < signature.length()) {
      typeClass.append(" implements ");
      typeClass.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    while (index < signature.length()) {
      typeClass.append(", ");
      typeClass.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    return typeClass.toString();
  }
  
  public static String toHexString(byte[] bytes) {
    StringBuilder buf = new StringBuilder();
    for (int i = 0; i < bytes.length; i++) {
      short b = byteToShort(bytes[i]);
      String hex = Integer.toHexString(b);
      if (b < 16)
        buf.append('0'); 
      buf.append(hex);
      if (i < bytes.length - 1)
        buf.append(' '); 
    } 
    return buf.toString();
  }
  
  public static byte typeOfMethodSignature(String signature) throws ClassFormatException {
    try {
      if (signature.charAt(0) != '(')
        throw new ClassFormatException("Invalid method signature: " + signature); 
      int index = signature.lastIndexOf(')') + 1;
      return typeOfSignature(signature.substring(index));
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid method signature: " + signature, e);
    } 
  }
  
  public static byte typeOfSignature(String signature) throws ClassFormatException {
    try {
      switch (signature.charAt(0)) {
        case 'B':
          return 8;
        case 'C':
          return 5;
        case 'D':
          return 7;
        case 'F':
          return 6;
        case 'I':
          return 10;
        case 'J':
          return 11;
        case 'L':
        case 'T':
          return 14;
        case '[':
          return 13;
        case 'V':
          return 12;
        case 'Z':
          return 4;
        case 'S':
          return 9;
        case '!':
        case '*':
        case '+':
          return typeOfSignature(signature.substring(1));
      } 
      throw new ClassFormatException("Invalid method signature: " + signature);
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid method signature: " + signature, e);
    } 
  }
  
  private static String typeParamTypesToString(String signature, boolean chopit) {
    StringBuilder typeParams = new StringBuilder("<");
    int index = 1;
    typeParams.append(typeParamTypeToString(signature.substring(index), chopit));
    index += unwrap(CONSUMER_CHARS);
    while (signature.charAt(index) != '>') {
      typeParams.append(", ");
      typeParams.append(typeParamTypeToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    wrap(CONSUMER_CHARS, index + 1);
    return typeParams.append(">").toString();
  }
  
  private static String typeParamTypeToString(String signature, boolean chopit) {
    int index = signature.indexOf(':');
    if (index <= 0)
      throw new ClassFormatException("Invalid type parameter signature: " + signature); 
    StringBuilder typeParam = new StringBuilder(signature.substring(0, index));
    index++;
    if (signature.charAt(index) != ':') {
      typeParam.append(" extends ");
      typeParam.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    while (signature.charAt(index) == ':') {
      index++;
      typeParam.append(" & ");
      typeParam.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    wrap(CONSUMER_CHARS, index);
    return typeParam.toString();
  }
  
  private static String typeSignaturesToString(String signature, boolean chopit, char term) {
    StringBuilder typeList = new StringBuilder(signature.substring(0, 1));
    int index = 1;
    if (signature.charAt(index) != term) {
      typeList.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    while (signature.charAt(index) != term) {
      typeList.append(", ");
      typeList.append(typeSignatureToString(signature.substring(index), chopit));
      index += unwrap(CONSUMER_CHARS);
    } 
    wrap(CONSUMER_CHARS, index + 1);
    return typeList.append(term).toString();
  }
  
  public static String typeSignatureToString(String signature, boolean chopit) throws ClassFormatException {
    wrap(CONSUMER_CHARS, 1);
    try {
      int index;
      int fromIndex;
      int n;
      int i;
      StringBuilder brackets;
      int bracketIndex;
      String type;
      StringBuilder stringBuilder1;
      int consumedChars;
      int j;
      int temp;
      switch (signature.charAt(0)) {
        case 'B':
          return "byte";
        case 'C':
          return "char";
        case 'D':
          return "double";
        case 'F':
          return "float";
        case 'I':
          return "int";
        case 'J':
          return "long";
        case 'T':
          index = signature.indexOf(';');
          if (index < 0)
            throw new ClassFormatException("Invalid type variable signature: " + signature); 
          wrap(CONSUMER_CHARS, index + 1);
          return compactClassName(signature.substring(1, index), chopit);
        case 'L':
          fromIndex = signature.indexOf('<');
          if (fromIndex < 0) {
            fromIndex = 0;
          } else {
            fromIndex = signature.indexOf('>', fromIndex);
            if (fromIndex < 0)
              throw new ClassFormatException("Invalid signature: " + signature); 
          } 
          i = signature.indexOf(';', fromIndex);
          if (i < 0)
            throw new ClassFormatException("Invalid signature: " + signature); 
          bracketIndex = signature.substring(0, i).indexOf('<');
          if (bracketIndex < 0) {
            wrap(CONSUMER_CHARS, i + 1);
            return compactClassName(signature.substring(1, i), chopit);
          } 
          fromIndex = signature.indexOf(';');
          if (fromIndex < 0)
            throw new ClassFormatException("Invalid signature: " + signature); 
          if (fromIndex < bracketIndex) {
            wrap(CONSUMER_CHARS, fromIndex + 1);
            return compactClassName(signature.substring(1, fromIndex), chopit);
          } 
          stringBuilder1 = (new StringBuilder(compactClassName(signature.substring(1, bracketIndex), chopit))).append("<");
          j = bracketIndex + 1;
          if (signature.charAt(j) == '+') {
            stringBuilder1.append("? extends ");
            j++;
          } else if (signature.charAt(j) == '-') {
            stringBuilder1.append("? super ");
            j++;
          } 
          if (signature.charAt(j) == '*') {
            stringBuilder1.append("?");
            j++;
          } else {
            stringBuilder1.append(typeSignatureToString(signature.substring(j), chopit));
            j = unwrap(CONSUMER_CHARS) + j;
            wrap(CONSUMER_CHARS, j);
          } 
          while (signature.charAt(j) != '>') {
            stringBuilder1.append(", ");
            if (signature.charAt(j) == '+') {
              stringBuilder1.append("? extends ");
              j++;
            } else if (signature.charAt(j) == '-') {
              stringBuilder1.append("? super ");
              j++;
            } 
            if (signature.charAt(j) == '*') {
              stringBuilder1.append("?");
              j++;
              continue;
            } 
            stringBuilder1.append(typeSignatureToString(signature.substring(j), chopit));
            j = unwrap(CONSUMER_CHARS) + j;
            wrap(CONSUMER_CHARS, j);
          } 
          j++;
          stringBuilder1.append(">");
          if (signature.charAt(j) == '.') {
            stringBuilder1.append(".");
            stringBuilder1.append(typeSignatureToString("L" + signature.substring(j + 1), chopit));
            j = unwrap(CONSUMER_CHARS) + j;
            wrap(CONSUMER_CHARS, j);
            return stringBuilder1.toString();
          } 
          if (signature.charAt(j) != ';')
            throw new ClassFormatException("Invalid signature: " + signature); 
          wrap(CONSUMER_CHARS, j + 1);
          return stringBuilder1.toString();
        case 'S':
          return "short";
        case 'Z':
          return "boolean";
        case '[':
          brackets = new StringBuilder();
          for (n = 0; signature.charAt(n) == '['; n++)
            brackets.append("[]"); 
          consumedChars = n;
          type = typeSignatureToString(signature.substring(n), chopit);
          temp = unwrap(CONSUMER_CHARS) + consumedChars;
          wrap(CONSUMER_CHARS, temp);
          return type + brackets.toString();
        case 'V':
          return "void";
      } 
      throw new ClassFormatException("Invalid signature: `" + signature + "'");
    } catch (StringIndexOutOfBoundsException e) {
      throw new ClassFormatException("Invalid signature: " + signature, e);
    } 
  }
  
  private static int unwrap(ThreadLocal<Integer> tl) {
    return ((Integer)tl.get()).intValue();
  }
  
  private static void wrap(ThreadLocal<Integer> tl, int value) {
    tl.set(Integer.valueOf(value));
  }
}
